/*
 * ouvej, determines what, when and where the next lesson will be
 * Copyright (C) 2003 Olivier Mehani <shtrom@ssji.net, olmehani@etu.utc.fr>
 * 
 * $Id: btree.c,v 1.21 2004/03/07 18:47:59 shtrom Exp $
 * binary tree handling functions 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h> /* for INT_MIN */
#include "ouvej.h"

bt_t *btree_make(ll_t *list)
{
	ll_t *cur_list=list;
	bt_t *btree=NULL;

/*	DPRINT("making binary tree from list at %p\n",list);
*/
	while(cur_list) {
		btree_add(&btree, cur_list->lesson);
/*		DPRINT("pointers: current=%p next=%p\n",cur_list, cur_list->next);
*/
		cur_list=cur_list->next;
	}
	
	return btree;
}

bt_t *btree_make_equilibrated(int size, lesson_t lessons[size])
{
	bt_t *btree;
	int half=size/2;

	if(size<1) {
		DPRINT("not enough entries (%d)\n",size);
		return NULL;
	}

	DPRINT("working with %d entries\n",size);
	
	DPRINT("pivot item at index %d\n",half);

	/*
	 * add the pivot as the root for the rest 
	 */
	btree=btree_new_node(lessons[half]);

	/*
	 * build the lower part tree and link it
	 */
	btree->earlier=btree_make_equilibrated(half,lessons);
	
	/*
	 * build the upper part and link it
	 *
	 * the number of elements to consider depends on whether the
	 * total number of elements was odd or even, we don't want to
	 * forget the last element... 
	 */
	btree->later=btree_make_equilibrated(size%2?(half):(half-1),&lessons[half+1]);
	
	return btree;
}

void btree_add(bt_t **btree, lesson_t l)
{
	bt_t *cur=*btree;

	DPRINT("adding lesson %s\n", l.name);

	if(!cur) {
		DPRINT("new tree\n");
		*btree=btree_new_node(l);
	} else {
		while(cur) {
			if(lesson_compare_beg(l,cur->lesson)<0) {
				DPRINT("lesson occurs earlier than %s\n",cur->lesson.name);
				if(cur->earlier==NULL) {
					cur->earlier=btree_new_node(l);
					cur=NULL;
				} else
					cur=cur->earlier;
			} else {
				/*
				 * Lesson with equal start hours (but different
				 * week sequence are stored in the later part
				 * for the current node
				 */
				DPRINT("lesson occurs later than %s\n",cur->lesson.name);
				if(cur->later==NULL) {
					cur->later=btree_new_node(l);
					cur=NULL;
				} else
					cur=cur->later;
			}	
		}
	}
}

bt_t *btree_new_node(lesson_t l)
{
	bt_t *node;

	DPRINT("allocating new node\n");

	node=malloc(sizeof(bt_t));
	if(!node) {
		ERROR("couldn't allocate memory\n");
		exit(1);
	}

	node->lesson=l;
	node->earlier=NULL;
	node->later=NULL;
	
	return node;
}

int btree_write(bt_t *btree, char *filename)
{
	lesson_t l;
	FILE *out; 

	DPRINT("writing to %s\n",filename);

	out=cache_wopen(filename);
	if(!out) {
		ERROR("can't open cache file %s for writing\n",filename);
		exit(1);
	}

	l=btree_get_smallest_and_remove(&btree);
	while (*l.name!=0) {
		DDO(lesson_show(l,stderr));
		fwrite(&l,sizeof(lesson_t),1,out);
		l=btree_get_smallest_and_remove(&btree);
	}

	fclose(out);
	
	return 0;
}


bt_t *btree_read(char *filename)
{
	bt_t *btree;
	lesson_t *llist;
	int num=0;
	
	llist=lesson_read(filename, &num);

	btree=btree_make_equilibrated(num, llist);
	
	free(llist);
	
	return btree;
}

lesson_t btree_get_smallest_and_remove(bt_t **btree)
{
	bt_t *tree;
	lesson_t l;

	memset(&l,0,sizeof(lesson_t));

	if(!btree)
		return l;

	tree=*btree;

	if(!tree)
		return l;

	if(tree->earlier) {
		return btree_get_smallest_and_remove(&tree->earlier);
	} else {
		*btree=tree->later;
		l=tree->lesson;
		free(tree);
		return l;
	}	
}

lesson_t btree_get_smallest(bt_t *btree)
{
	lesson_t l;

	memset(&l,0,sizeof(lesson_t));

	if(!btree)
		return l;

	if(btree->earlier) {
		return btree_get_smallest(btree->earlier);
	} else {
		btree=btree->later;
		l=btree->lesson;
		return l;
	}	
}

void btree_show(bt_t *btree)
{
	/*
	 * This is a dirty debug fonction
	 */
	static int inc_level=0;
	int i;
	

	if(!inc_level)
		printf("ROOT\n");


	if(btree) {
		inc_level++;
		for(i=0;i<inc_level-1;i++)
			printf(" |   ");
		printf(" +-> ");

		lesson_show(btree->lesson,stdout);

		btree_show(btree->earlier);
		btree_show(btree->later);

		inc_level--;
	}

	if(!inc_level)
		printf("END\n");
}

lesson_t btree_find_next(bt_t *btree, lesson_t l)
{
	bt_t *cur=btree, *last=NULL;
	int min_diff=INT_MIN, temp;
	lesson_t first, second;

	l.name[0]=0;

	DPRINT("seeking lesson ");
	DDO(lesson_show(l,stderr));

	if(lesson_get_seq(&l)==W_N)
		return l;

	while(cur) {
		temp=lesson_compare_end(l,cur->lesson);
		if(temp<0) {
			DPRINT("lesson occurs earlier (%d min) than ",temp);
			DDO(lesson_show(cur->lesson,stderr));
			
			if(!lesson_match_week_seq(l,cur->lesson)) {
				DPRINT("current lesson doesn't occur this week, splitting\n");
				first=btree_find_next(cur->earlier,l);
				second=btree_find_next(cur->later,l);
				
				if(first.name[0] && second.name[0]) { 
					if(lesson_compare_beg(first,second)<0)
						return first;
					else
						return second;
				} else if(first.name[0]) {
					return first;
				} else if(second.name[0]) {
					return second;
				} else {
					return last->lesson;
				}

				return l;
			} else {
				if(min_diff<temp) {
					DPRINT("closer than the closest (%d<%d)\n",min_diff,temp);
					min_diff=temp;
					last=cur;
				}
				cur=cur->earlier;
				if(!cur) {
					if(last)
						return last->lesson;
					else
						return l;
				}
			}

		} else {
			DPRINT("lesson occurs later (%d min) than ", temp);
			DDO(lesson_show(cur->lesson,stderr));
			
/*			if(last)
				return last->lesson;
			if(!cur->later)
				return cur->lesson;
*/
			cur=cur->later;
			if (!cur) {
				if(last)
					return last->lesson;
				else
					return l;
			}
		}	
	}

	return l;
}
